/****************************************************************
                                                                *
        This file is just the front-end application for		*
	configuring the driver.  Nothing too special.		*
                                                                *
*****************************************************************/

#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/sockios.h>
#include <linux/if.h>
#include <linux/string.h>
#include <stdio.h>
#include <unistd.h>
#ifdef CONFIG_NET_RADIO  /* If wireless extension exist in the kernel */
/* #include <linux/wireless.h>  don't include for now ... */
#endif /* CONFIG_NET_RADIO */

#include "lld.h"
#include "proxcfg.h"
#define SPACE printf("  ")
#define TAB printf("\t")
#define RETURN printf("\n")

struct rl2_ioctl_info *ioc;
struct ifreq ifr;
unsigned char CardMode;
void PrintUsage(void);
void atomac(unsigned_8 * string, unsigned_8 * mac);

int main(int argc, char *argv[])
{
	int s,i,j,k;
	unsigned int resetLLD=CLEAR;
	unsigned char data_string[21];
	unsigned int data_val;

	s=socket(AF_INET, SOCK_DGRAM, 0);
	ioc = (struct rl2_ioctl_info *) &ifr.ifr_data;

/**************************************************************** 
   Print usage if called for
****************************************************************/   
	if(argc == 1 || !strcmp(argv[1], "--help")) {
		PrintUsage();
		return 0;
	}
	
/**************************************************************** 
   Open a socket to the driver
****************************************************************/   
	if(s==-1) {
	  perror("proxnet: socket");
	  exit(1);
	}

/**************************************************************** 
   Check for some default stuff.  We take eth0 as a default
   or print out info if you didn't try to set a parameter.
****************************************************************/   
	if(!strcmp(argv[1],"dev")) {
    	  if(argc >= 3) {
	    strcpy(ifr.ifr_name,argv[2]);
	    if(argc == 3) PrintUsage();
	    i=3;
	  }
	  else {
  	    strcpy(ifr.ifr_name, "eth0");
	    i=2;
	    PrintUsage();
	  }
	}
	else {
  	   strcpy(ifr.ifr_name, "eth0");
	   i=1;
	}

/**************************************************************** 
   This parses the arguments and makes the call to set the
   parameter via dev ioctl to lls.c.  In general, LLDReset we
   try to call only once so we note if it's needed and call once
   at the end after parsing all the parameters.
****************************************************************/   
	for(;i<argc;i++) {
		if(!strcmp(argv[i],"ping")) {
		  if (++i >= argc) {
		  	PrintUsage();
		  	exit(1); }
		  ioc->cmd = SEND_PROX_PING;
		  ioc->data = data_string;
		  atomac(argv[i], data_string);
		  if(ioctl(s,SIOCDEVPRIVATE,&ifr) == -1) {
		    perror("proxnet: ioctl");
		    PrintUsage();
		    exit(1);
		  }
		  printf("Check for responses:  tail /var/log/messages\n");
		}
		else if(!strcmp(argv[i],"masters")) {
		  struct MastersList currlist;
		  if (++i >= argc) {
		  	PrintUsage();
		  	exit(1); }
		  ioc->cmd = START_AP_SEARCH;
		  ioc->data = data_string;
		  if(ioctl(s,SIOCDEVPRIVATE,&ifr) == -1) {
		    perror("net: ioctl");
		    PrintUsage();
		    exit(1);
		  }
		 for (k=0; k<atoi(argv[i])/5; k++) {
		  sleep(5);
		  ioc->data = (unsigned_8 *) &currlist;
		  ioc->cmd = AP_LIST;
		  ioctl(s,SIOCDEVPRIVATE,&ifr);
		  printf("Time: %d\n", (k+1)*5);
		  if (currlist.ChSubChannel) {
		    printf("Name:%s  MAC:", currlist.Name );
		    for(j=0;j<5;j++)
		  	  printf("%2.2x:", currlist.Address[j]);
		    printf("%2.2x", currlist.Address[j]);
		    printf("%d, %d\n", currlist.ChSubChannel&0xf, currlist.ChSubChannel);
		  }
		  else printf("No masters found.\n");
		 }
		  ioc->data = (unsigned_8 *) &currlist;
		  ioc->cmd = STOP_AP_SEARCH;
		  if(ioctl(s,SIOCDEVPRIVATE,&ifr) == -1) {
		    perror("proxnet: ioctl");
		    PrintUsage();
		    exit(1);
		  }
		}
		else {
		  PrintUsage();
		  exit(1);
		}
	}
}


void PrintUsage(void)
{
	printf("Usage:\tproxnet\t--help\n");
	printf("\tproxnet\t[dev {ifname}] [ping <wireless mac address>] [masters <search time>]\n");
}


void atomac(unsigned_8 * string, unsigned_8 * mac)
{
		  unsigned_8 index, j, k, digit;
		  for (j=0; j<6; j++) {
		    mac[j]=0;
		    for(k=0;k<2;k++) {
			index=2*j+k;
			if       ((digit = *(string+index)-'0') < 10);
			else if (((digit = *(string+index)-'a'+10) >=0xa) && (digit <= 0xf));
			else if (((digit = *(string+index)-'A'+10) >=0xa) && (digit <= 0xf));
			else digit=0;
			mac[j] += ( (digit&=0x0f) << ((1-k)*4));
		    }
		  }
}
